<?php
/**
* @package   awl
* @subpackage   AwlDatabase
* @author    Andrew McMillan <andrew@morphoss.com>
* @copyright Morphoss Ltd
* @license   http://www.gnu.org/licenses/lgpl-3.0.txt  GNU LGPL version 3 or later
* @compatibility Requires PHP 5.1 or later
*/

require_once('AwlQuery.php');

/**
* Database upgrader class and associated functions
*
* This subpackage provides some functions that are useful around database
* schema creation and changes.
*
*/

/**
* The AwlUpgrader Class.
*
* This class updates an Awl database to a newer schema version.
*
*
* @package   awl
*/
class AwlUpgrader
{
 /**
  * Constructor
  * @return The AwlUpgrader object
  */
  function __construct() {
    return $this;
  }
}

